
--============================================================================= 
--
-- Mode Controller
--
-- Toggles through choosing 1 out of 4 possible LED sequence modes
--
--============================================================================= 

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

entity control_mux is
    port (
        clk : in std_logic;
        user_btn : in std_logic;
        led_in : in std_logic_vector(31 downto 0);
        led_out : out std_logic_vector(7 downto 0);
        sel : out std_logic_vector(3 downto 0)
    );
end entity;

architecture behavioural of control_mux is

    signal sel_reg : std_logic_vector(3 downto 0) := "0001";
    signal sel_reg_nxt : std_logic_vector(3 downto 0) := "0001";

    signal key_db : std_logic := '0';
    signal key_db_nxt : std_logic := '0';
    signal key_delay : std_logic := '0';

    signal dbcnt : integer range 0 to 63 := 0;
    signal dbcnt_nxt : integer range 0 to 63 := 0;

    signal key_fall : std_logic := '0';
begin

    -------------------------------------------------------------------------------
    -- register
    -------------------------------------------------------------------------------
    proc_name : process (clk)
    begin

        if rising_edge(clk) then
            sel_reg <= sel_reg_nxt;
            key_db <= key_db_nxt;
            dbcnt <= dbcnt_nxt;

            key_delay <= key_db;
        end if;
    end process proc_name;
    -------------------------------------------------------------------------------
    -- debounce key
    -------------------------------------------------------------------------------

    process (key_db, dbcnt, user_btn)
    begin
        dbcnt_nxt <= dbcnt;
        key_db_nxt <= key_db;
        if (user_btn = key_db) then
            dbcnt_nxt <= 0;
        else
            dbcnt_nxt <= dbcnt + 1;
        end if;

        if (dbcnt = 63) then
            key_db_nxt <= user_btn;
        end if;
    end process;

    -------------------------------------------------------------------------------
    -- falling edge detect
    -------------------------------------------------------------------------------
    key_fall <= key_delay and not (key_db);
    -----------------------------------------------------------
    -- toggle through states
    -------------------------------------------------------------------------------

    process (key_fall, sel_reg)
    begin
        sel_reg_nxt <= sel_reg(3 downto 0);
        if key_fall = '1' then
            sel_reg_nxt(3 downto 0) <= sel_reg(2 downto 0) & sel_reg(3);
        end if;
    end process;
    -------------------------------------------------------------------------------
    -- Pass LED sequence to LEDs depending on mode
    -------------------------------------------------------------------------------
    process (sel_reg, led_in)
    begin
        case sel_reg is
            when "0001" => led_out <= led_in (7 downto 0);
            when "0010" => led_out <= led_in (15 downto 8);
            when "0100" => led_out <= led_in (23 downto 16);
            when "1000" => led_out <= led_in (31 downto 24);
            when others => led_out <= "00000000";
        end case;
    end process;

    sel <= sel_reg;

end architecture;